package ca.ntro.app.world2d;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import ca.ntro.app.fx.controls.ResizableWorld2dCanvasFx;
import ca.ntro.app.fx.controls.World2dMouseEventFx;
import ca.ntro.app.models.ModelValue;
import ca.ntro.core.NtroCore;
import ca.ntro.ntro_app_fx_impl.world2d.region_tree.AnonymousRegion2dNtro;
import ca.ntro.ntro_core_impl.NtroCoreImpl;

public abstract class World2dFx implements ModelValue {

	private static final double INTERSECT_EPSILON = 5;

	public double width = 640;
	public double height = 360;

	public transient List<Object2dFx<?>> objects = new ArrayList<>();
	
	private transient boolean clipAtViewport = true;

	public double getWidth() {
		return width;
	}

	public void setWidth(double width) {
		this.width = width;
	}

	public double getHeight() {
		return height;
	}

	public void setHeight(double height) {
		this.height = height;
	}

	public List<Object2dFx<?>> getObjects() {
		return objects;
	}

	public void setObjects(List<Object2dFx<?>> objects) {
		this.objects = objects;
	}
	
	public boolean getClipAtViewport() {
		return clipAtViewport;
	}

	public void setClipAtViewport(boolean clipAtViewport) {
		this.clipAtViewport = clipAtViewport;
	}

	public World2dFx() {
        initialize();
	}
	
	
	protected abstract void initialize();
	protected abstract void onMouseEventNotConsumed(World2dMouseEventFx mouseEvent);
	
	public void initializeFx() {
		for(Object2dFx<?> object2d: objects) {
			object2d.initializeFx();
		}
	}
	

	/**
	 * Adds an object to the world2d
	 * Calls initialize of the object
	 * @param object2d
	 */
	@SuppressWarnings("unchecked")
	public <W extends World2dFx> void addObject2d(Object2dFx<W> object2d) {
		NtroCoreImpl.threads().runOnMainThread(() -> {
			object2d.setWorld2d((W) this);
			objects.add(object2d);
			object2d.initialize();
		});
	}

	/**
	 * Calls drawOn for each object added to the world2d
	 * @param canvas
	 */
	public void drawOn(ResizableWorld2dCanvasFx canvas) {

		AnonymousRegion2dNtro viewport = new AnonymousRegion2dNtro(canvas.getViewportTopLeftX(), 
				                                                   canvas.getViewportTopLeftY(), 
				                                                   canvas.getViewportWidth(), 
				                                                   canvas.getViewportHeight());
		
		// TODO: find objects that intersect with 
		//       the viewport in log(n) using
		//       a RegionTree
		
		canvas.drawOnWorld(gc -> {
			for(Object2dFx<?> object2d: objects) {
				if(!clipAtViewport ||
						object2d.intersectsWith(viewport, INTERSECT_EPSILON)) {

					object2d.drawOnWorld(gc);
				}
			}
		});
	}

	public void onTimePasses(double secondsElapsed) {
		for(Object2dFx<?> objet2d: objects) {
			objet2d.onTimePasses(secondsElapsed);
		}
	}

	public void copyDataFrom(World2dFx world2d) {
		for(int i = 0; i < objects.size(); i++) {
			world2d.copyDataTo(i, objects.get(i));
		}
	}

	protected void copyDataTo(int i, Object2dFx<?> object2d) {
		if(i < objects.size()) {
			object2d.copyDataFrom(objects.get(i));
		}
	}

	public void dispatchMouseEvent(World2dMouseEventFx world2dMouseEventFx) {
        double worldX = world2dMouseEventFx.worldX();
        double worldY = world2dMouseEventFx.worldY();

        boolean consumed = false;
        for(Object2dFx<?> object : getObjects()) {

            if(object.collidesWith(worldX - INTERSECT_EPSILON, 
            		               worldY - INTERSECT_EPSILON, 
            		               2 * INTERSECT_EPSILON, 
            		               2 * INTERSECT_EPSILON)) {

                consumed = consumed || object.onMouseEvent(world2dMouseEventFx);
            }
        }

        if(!consumed) {
            onMouseEventNotConsumed(world2dMouseEventFx);
        }
	}

	public void removeObject(String objectId) {
		NtroCore.threads().runOnMainThread(() -> {
			Integer indexOfObjectToRemove = objectIndexById(objectId);

			if(indexOfObjectToRemove != null) {
				getObjects().remove((int) indexOfObjectToRemove);
			}
		});
	}

	public void removeObjectsWithIdIn(Set<String> idsOfObjectsToRemove) {
		NtroCore.threads().runOnMainThread(() -> {
			for(String idOfObjectToRemove : idsOfObjectsToRemove) {
				Integer indexOfObjectToRemove = objectIndexById(idOfObjectToRemove);
				
				if(indexOfObjectToRemove != null) {
						getObjects().remove((int) indexOfObjectToRemove);
				}
			}
		});
	}

	public void removeObjectsWithIdNotIn(Set<String> idsOfObjectToKeep) {
		NtroCore.threads().runOnMainThread(() -> {
			int i = 0;
			while(i < getObjects().size()) {
				Object2dFx<?> candidate = getObjects().get(i);

				if(candidate != null
						&& candidate.id() != null
						&& !(idsOfObjectToKeep.contains(candidate.id()))) {
					
					getObjects().remove(i);

				}else {
					i++;
				}
			}
		});
	}

	private Integer objectIndexById(String id) {
		Integer index = null;

		for(int i = 0; i < getObjects().size(); i++) {
			Object2dFx<?> candidate = getObjects().get(i);

			if(candidate != null
					&& candidate.id() != null
					&& candidate.id().equals(id)) {

				index = i;
				break;
			}
		}

		return index;
	}

	public Object2dFx<?> objectById(String id) {
		Object2dFx<?> result = null;
		Integer objectIndex = objectIndexById(id);
		
		if(objectIndex != null) {
			result = getObjects().get(objectIndex);
		}

		return result;
	}
}